"""Mapping files for writing to the registries."""
from .special import Special

DEVICE_WRITE_HCMAP = {
    "hc0": {
        "ambient_requested": [
            Special(float, Special.FORMULA),
            [
                "'A_191' if {0} else 'A_210'",
                [[bool, "D_242"]],
            ],
        ],
    },
    "hc1": {
        "on": [bool, "D_212"],
        "ambient_requested": [
            Special(float, Special.FORMULA),
            [
                "'A_219' if {0} else 'A_215'",
                [[bool, "D_245"]],
            ],
        ],
        "pad": {
            "state": [int, "I_15"],  # 0 - Permanently Off, 1 - Scheduled Off, 2 - On
        },
        "control_curve_heating": {
            "setpoint_a_outside": [float, "A_101"],
            "setpoint_a_requested": [float, "A_106"],
            "setpoint_b_outside": [float, "A_102"],
            "setpoint_b_requested": [float, "A_107"],
        },
        "control_curve_cooling": {
            "setpoint_a_outside": [float, "A_314"],
            "setpoint_a_requested": [float, "A_315"],
            "setpoint_b_outside": [float, "A_316"],
            "setpoint_b_requested": [float, "A_317"],
        },
    },
    "hc2": {
        "on": [bool, "D_216"],
        "ambient_requested": [
            Special(float, Special.FORMULA),
            [
                "'A_225' if {0} else 'A_221'",
                [[bool, "D_248"]],
            ],
        ],
        "pad": {
            "state": [int, "I_6"],
        },
        "control_curve_heating": {
            "setpoint_a_outside": [float, "A_108"],
            "setpoint_a_requested": [float, "A_84"],
            "setpoint_b_outside": [float, "A_109"],
            "setpoint_b_requested": [float, "A_85"],
        },
        "control_curve_cooling": {
            "setpoint_a_outside": [float, "A_330"],
            "setpoint_a_requested": [float, "A_331"],
            "setpoint_b_outside": [float, "A_332"],
            "setpoint_b_requested": [float, "A_333"],
        },
    },
    "hc3": {
        "on": [bool, "D_220"],
        "ambient_requested": [
            Special(float, Special.FORMULA),
            [
                "'A_231' if {0} else 'A_227'",
                [[bool, "D_251"]],
            ],
        ],
        "pad": {
            "state": [int, "I_227"],
        },
        "control_curve_heating": {
            "setpoint_a_outside": [float, "A_113"],
            "setpoint_a_requested": [float, "A_86"],
            "setpoint_b_outside": [float, "A_114"],
            "setpoint_b_requested": [float, "A_87"],
        },
        "control_curve_cooling": {
            "setpoint_a_outside": [float, "A_346"],
            "setpoint_a_requested": [float, "A_347"],
            "setpoint_b_outside": [float, "A_348"],
            "setpoint_b_requested": [float, "A_349"],
        },
    },
    "hc4": {
        "on": [bool, "D_50"],
        "ambient_requested": [
            Special(float, Special.FORMULA),
            [
                "'A_238' if {0} else 'A_233'",
                [[bool, "D_254"]],
            ],
        ],
        "pad": {
            "state": [int, "I_230"],
        },
        "control_curve_heating": {
            "setpoint_a_outside": [float, "A_122"],
            "setpoint_a_requested": [float, "A_120"],
            "setpoint_b_outside": [float, "A_88"],
            "setpoint_b_requested": [float, "A_121"],
        },
        "control_curve_cooling": {
            "setpoint_a_outside": [float, "A_362"],
            "setpoint_a_requested": [float, "A_363"],
            "setpoint_b_outside": [float, "A_364"],
            "setpoint_b_requested": [float, "A_365"],
        },
    },
    "hc5": {
        "on": [bool, "D_51"],
        "ambient_requested": [
            Special(float, Special.FORMULA),
            [
                "'A_247' if {0} else 'A_240'",
                [[bool, "D_257"]],
            ],
        ],
        "pad": {
            "state": [int, "I_239"],
        },
        "control_curve_heating": {
            "setpoint_a_outside": [float, "A_387"],
            "setpoint_a_requested": [float, "A_388"],
            "setpoint_b_outside": [float, "A_389"],
            "setpoint_b_requested": [float, "A_390"],
        },
        "control_curve_cooling": {
            "setpoint_a_outside": [float, "A_379"],
            "setpoint_a_requested": [float, "A_380"],
            "setpoint_b_outside": [float, "A_381"],
            "setpoint_b_requested": [float, "A_382"],
        },
    },
    "hc6": {
        "on": [bool, "D_52"],
        "ambient_requested": [
            Special(float, Special.FORMULA),
            [
                "'A_249' if {0} else 'A_277'",
                [[bool, "D_259"]],
            ],
        ],
        "pad": {
            "state": [int, "I_210"],
        },
        "control_curve_heating": {
            "setpoint_a_outside": [float, "A_401"],
            "setpoint_a_requested": [float, "A_402"],
            "setpoint_b_outside": [float, "A_403"],
            "setpoint_b_requested": [float, "A_404"],
        },
        "control_curve_cooling": {
            "setpoint_a_outside": [float, "A_405"],
            "setpoint_a_requested": [float, "A_406"],
            "setpoint_b_outside": [float, "A_407"],
            "setpoint_b_requested": [float, "A_408"],
        },
    },
    "pool": {
        "on": [bool, "D_238"],
        "temp_requested": [float, "A_210"],
    },
}

DEVICE_WRITE_MAP = {
    "hp_power_state": [bool, "D_3"],
    "hp_function": [int, "I_51"],  # 0: heating, #1: cooling, #2: auto (Write)
    "season": {
        # Based on Average Temperature, set manual_set then winter if manual.
        "manual_set": [bool, "I_50"],  # True if Manual Set, False if Auto
        "winter": [bool, "D_24"],  # True is Winter, False is Summer
        "winter_temp": [float, "A_82"],  # -20.0 to 40.0 Below Winter Mode
        "summer_temp": [float, "A_83"],  # -20.0 to 40.0 Above Summer Mode
    },
    "control_curve_heating": {  # -99.9 to 99.9
        "setpoint_a_outside": [float, "A_35"],
        "setpoint_a_requested": [float, "A_37"],
        "setpoint_b_outside": [float, "A_36"],
        "setpoint_b_requested": [float, "A_38"],
    },
    "control_curve_cooling": {  # -99.9 to 99.9
        "setpoint_a_outside": [float, "A_47"],
        "setpoint_a_requested": [float, "A_48"],
        "setpoint_b_outside": [float, "A_49"],
        "setpoint_b_requested": [float, "A_50"],
    },
    "domestic_hot_water": {
        "state": [bool, "D_29"],
        "required_temp": [float, "A_129"],  # Range min_temp to max_temp
    },
    "heating_circuits": DEVICE_WRITE_HCMAP,
}



"""Test Configuration for various tests."""
import json
import os

VALID_LOGIN = {
    "uname": "user name@email",
    "upwd": "hash%pass or",
    "token": "9jdhhs78dodlosd",
}

GENERAL_ERROR_RESPONSE = "User not logged in"


def load_fixture(filename):
    """Load a JSON fixture for testing."""
    try:
        path = os.path.join(os.path.dirname(__file__), "fixtures", filename)
        with open(path, encoding="utf-8") as fptr:
            return fptr.read()
    except OSError:
        return None


class ConnectionMock:
    """Mock the Connection Class to return what we want."""

    def __init__(self, api_version: str = "v1", use_mt: bool = False):
        """Initialize the Connection Mock."""
        self.__api_version = api_version

        if api_version == "v1":
            if use_mt:
                self.__subfolder = "mt/"
            else:
                self.__subfolder = ""
        else:
            self.__subfolder = "newapi/"

        self.__devices: dict = {}

    def connect(self, role="400"):
        """Mock Connect, Return Connect Success/ Failures."""
        if self.__api_version == "v1":
            result = json.loads(load_fixture(f"{self.__subfolder}login_success.json"))
        else:
            result = json.loads(load_fixture(f"{self.__subfolder}modules.json"))

        result["role"] = role
        self.__devices = {}
        for module in result["modules"]:
            for unit in module["config"]:
                device_id = module["id"] + "_" + str(unit["mb_addr"])
                self.__devices[device_id] = {"data": None}

        return result

    def get_device_info(self, module_id, unit_id):
        """Return the Device Information from fixtures."""
        info = json.loads(
            load_fixture(f"{self.__subfolder}pumpinfo_{module_id}_{unit_id}.json")
        )

        return info

    def get_device_data(self, module_id, unit_id, last_update_time=None):
        """Return Device Data from fixtures, fixed for Controller Test."""
        if last_update_time is None:
            last_update_time = "0"

        device_id = module_id + "_" + unit_id
        if self.__devices[device_id]["data"] is None or last_update_time != 0:
            data = json.loads(
                load_fixture(
                    f"{self.__subfolder}pumpdata_{module_id}_{unit_id}_{last_update_time}.json"
                )
            )
            if "varfile_mt1_config1" in data["data"]:
                var_data = "varfile_mt1_config1"
            elif "varfile_mt1_config2" in data["data"]:
                var_data = "varfile_mt1_config2"
            else:
                var_data = "varFileData"

            data["data"]["varData"] = data["data"][var_data]
            del data["data"][var_data]
            self.__devices[device_id]["data"] = data
        else:
            data = self.__devices[device_id]["data"]

        return data

    def set_device_data(
        self, module_id: str, unit_id: str, reg: str, value: str
    ) -> bool:
        """Update the Device Data and Return."""
        device_id = module_id + "_" + unit_id
        self.__devices[device_id]["data"]["data"]["varData"][unit_id.zfill(3)][
            reg
        ] = value
        return True



"""Test the Configuration data in const."""
from unittest.mock import patch

from aiohttp import ClientSession
from masterthermconnect import MasterthermController

from .conftest import VALID_LOGIN, ConnectionMock


async def test_setup():
    """Test the Setup."""
    controller = MasterthermController(
        VALID_LOGIN["uname"], VALID_LOGIN["upwd"], ClientSession()
    )
    mockconnect = ConnectionMock()

    with patch(
        "masterthermconnect.api.MasterthermAPI.connect",
        return_value=mockconnect.connect(),
    ), patch(
        "masterthermconnect.api.MasterthermAPI.get_device_info",
        side_effect=mockconnect.get_device_info,
    ), patch(
        "masterthermconnect.api.MasterthermAPI.get_device_data",
        side_effect=mockconnect.get_device_data,
    ):
        assert await controller.connect() is True
        assert await controller.refresh() is True

    # Verify basic setup.
    data = controller.get_device_data("1234", "1")
    assert data["hp_power_state"]
    assert data["control_curve_heating"]["setpoint_a_outside"] == 20.0


async def test_hc0():
    """Test basic setup HC0 only no optional modules."""
    controller = MasterthermController(
        VALID_LOGIN["uname"], VALID_LOGIN["upwd"], ClientSession()
    )
    mockconnect = ConnectionMock(api_version="v2")

    with patch(
        "masterthermconnect.api.MasterthermAPI.connect",
        return_value=mockconnect.connect(),
    ), patch(
        "masterthermconnect.api.MasterthermAPI.get_device_info",
        side_effect=mockconnect.get_device_info,
    ), patch(
        "masterthermconnect.api.MasterthermAPI.get_device_data",
        side_effect=mockconnect.get_device_data,
    ):
        assert await controller.connect() is True
        assert await controller.refresh() is True

    # Verify Setup with HC0 only
    data = controller.get_device_data("10021", "2")
    assert data["hp_power_state"]
    assert data["heating_circuits"]["hc0"]["enabled"]
    assert data["heating_circuits"]["hc0"]["pad"]["enabled"]


async def test_hc1_no_thermostat():
    """Test no thermostat installed in HC1."""
    controller = MasterthermController(
        VALID_LOGIN["uname"], VALID_LOGIN["upwd"], ClientSession()
    )
    mockconnect = ConnectionMock(api_version="v1")

    with patch(
        "masterthermconnect.api.MasterthermAPI.connect",
        return_value=mockconnect.connect(),
    ), patch(
        "masterthermconnect.api.MasterthermAPI.get_device_info",
        side_effect=mockconnect.get_device_info,
    ), patch(
        "masterthermconnect.api.MasterthermAPI.get_device_data",
        side_effect=mockconnect.get_device_data,
    ):
        assert await controller.connect() is True
        assert await controller.refresh() is True

    # Verify Setup with HC0 only
    data = controller.get_device_data("1234", "1")
    assert data["hp_power_state"]
    assert data["heating_circuits"]["hc1"]["enabled"]
    assert data["heating_circuits"]["hc1"]["ambient_requested"] == 20.2
    assert "pad" not in data["heating_circuits"]["hc1"]


async def test_hcx_thermostat():
    """Test Thermostat installed in HC."""
    controller = MasterthermController(
        VALID_LOGIN["uname"], VALID_LOGIN["upwd"], ClientSession()
    )
    mockconnect = ConnectionMock(api_version="v2")

    with patch(
        "masterthermconnect.api.MasterthermAPI.connect",
        return_value=mockconnect.connect(),
    ), patch(
        "masterthermconnect.api.MasterthermAPI.get_device_info",
        side_effect=mockconnect.get_device_info,
    ), patch(
        "masterthermconnect.api.MasterthermAPI.get_device_data",
        side_effect=mockconnect.get_device_data,
    ):
        assert await controller.connect() is True
        assert await controller.refresh() is True

    # Verify Setup with HC0 only
    data = controller.get_device_data("10021", "1")
    assert data["hp_power_state"]
    assert data["heating_circuits"]["hc2"]["enabled"]
    assert data["heating_circuits"]["hc2"]["ambient_requested"] == 20.1


async def test_hcx_loop_type():
    """Test HCx Loop Type."""
    controller = MasterthermController(
        VALID_LOGIN["uname"], VALID_LOGIN["upwd"], ClientSession()
    )
    mockconnect = ConnectionMock(api_version="v1")

    with patch(
        "masterthermconnect.api.MasterthermAPI.connect",
        return_value=mockconnect.connect(),
    ), patch(
        "masterthermconnect.api.MasterthermAPI.get_device_info",
        side_effect=mockconnect.get_device_info,
    ), patch(
        "masterthermconnect.api.MasterthermAPI.get_device_data",
        side_effect=mockconnect.get_device_data,
    ):
        assert await controller.connect() is True
        assert await controller.refresh() is True

    # Verify Setup with HC1 and HC2
    data = controller.get_device_data("1234", "1")
    assert data["heating_circuits"]["hc1"]["loop_type"] == 3
    assert data["heating_circuits"]["hc2"]["loop_type"] == 3


async def test_set_ambient_requested():
    """Test HC1 Ambient Requested."""
    controller = MasterthermController(
        VALID_LOGIN["uname"], VALID_LOGIN["upwd"], ClientSession()
    )
    mockconnect = ConnectionMock(api_version="v1")

    with patch(
        "masterthermconnect.api.MasterthermAPI.connect",
        return_value=mockconnect.connect(),
    ), patch(
        "masterthermconnect.api.MasterthermAPI.get_device_info",
        side_effect=mockconnect.get_device_info,
    ), patch(
        "masterthermconnect.api.MasterthermAPI.get_device_data",
        side_effect=mockconnect.get_device_data,
    ), patch(
        "masterthermconnect.api.MasterthermAPI.set_device_data",
        side_effect=mockconnect.set_device_data,
    ):
        assert await controller.connect() is True
        assert await controller.refresh() is True

        data = controller.get_device_data("1234", "1")
        assert await controller.set_device_data_item(
            "1234", "1", "heating_circuits.hc1.ambient_requested", 50.1
        )

        # Verfiy we update internally.
        registers = controller.get_device_registers("1234", "1")
        assert registers["A_215"] == "50.1"

        await controller.refresh(full_load=True)
        data = controller.get_device_data_item(
            "1234", "1", "heating_circuits.hc1.ambient_requested"
        )
        assert data == 50.1


async def test_season_set():
    """Test Season Sets correctly."""
    controller = MasterthermController(
        VALID_LOGIN["uname"], VALID_LOGIN["upwd"], ClientSession()
    )
    mockconnect = ConnectionMock(api_version="v1")

    with patch(
        "masterthermconnect.api.MasterthermAPI.connect",
        return_value=mockconnect.connect(),
    ), patch(
        "masterthermconnect.api.MasterthermAPI.get_device_info",
        side_effect=mockconnect.get_device_info,
    ), patch(
        "masterthermconnect.api.MasterthermAPI.get_device_data",
        side_effect=mockconnect.get_device_data,
    ), patch(
        "masterthermconnect.api.MasterthermAPI.set_device_data",
        side_effect=mockconnect.set_device_data,
    ):
        assert await controller.connect()
        assert await controller.refresh()

        # Test Winter Set
        assert await controller.set_device_data_item(
            "1234", "1", "season.manual_set", True
        )
        assert await controller.set_device_data_item("1234", "1", "season.winter", True)

        assert await controller.refresh(full_load=True)
        season = controller.get_device_data_item("1234", "1", "season.mode")
        assert season == "winter"

        # Test Summer Set
        assert await controller.set_device_data_item(
            "1234", "1", "season.manual_set", True
        )
        assert await controller.set_device_data_item(
            "1234", "1", "season.winter", False
        )

        assert await controller.refresh(full_load=True)
        season = controller.get_device_data_item("1234", "1", "season.mode")
        assert season == "summer"

        # Test Auto Winter Set
        assert await controller.set_device_data_item(
            "1234", "1", "season.manual_set", False
        )
        assert await controller.set_device_data_item("1234", "1", "season.winter", True)

        assert await controller.refresh(full_load=True)
        season = controller.get_device_data_item("1234", "1", "season.mode")
        assert season == "auto-winter"



"""Mastertherm API Client Class, handle integration."""

import asyncio
import logging
import time

from datetime import datetime, timedelta
from hashlib import sha1
from json.decoder import JSONDecodeError
from urllib.parse import urljoin, quote_plus

from natsort import natsorted

from aiohttp import ClientSession, ClientConnectionError, ContentTypeError

from .const import (
    APP_CLIENTINFO,
    APP_CLIENTINFO_NEW,
    DATE_FORMAT,
    SUPPORTED_ROLES,
    SUPPORTED_API_VERSIONS,
    URL_BASE,
    URL_BASE_NEW,
    URL_LOGIN,
    URL_LOGIN_NEW,
    URL_MODULES_NEW,
    URL_PUMPDATA,
    URL_PUMPDATA_NEW,
    URL_PUMPINFO,
    URL_PUMPINFO_NEW,
    URL_POSTUPDATE,
    URL_POSTUPDATE_NEW,
)
from .exceptions import (
    MasterthermAuthenticationError,
    MasterthermConnectionError,
    MasterthermUnsupportedRole,
    MasterthermUnsupportedVersion,
    MasterthermTokenInvalid,
    MasterthermResponseFormatError,
    MasterthermPumpError,
    MasterthermServerTimeoutError,
)

_LOGGER: logging.Logger = logging.getLogger(__name__)


class MasterthermAPI:
    """API Handler for the Mastertherm API."""

    def __init__(
        self,
        username: str,
        password: str,
        session: ClientSession,
        api_version: str,
    ) -> None:
        """Initialise the Mastertherm API Client.

        Args:
            username: The Login Username
            password: The Login Password
            session: an aiohttp client session
            api_version: The version of the API, mainly the host
                "v1"  : Original version, data response in varfile_mt1_config1 or 2
                "v2"  : New version since 2022 response in varFileData

        Returns:
            The MasterthermAPI object

        Raises:
            MasterthermUnsupportedVersion: API Version is not supported.

        """
        if api_version not in SUPPORTED_API_VERSIONS:
            raise MasterthermUnsupportedVersion(
                "-1", f"Unsupported Version {api_version}"
            )

        self.__session = session
        self.__api_version = api_version
        self.__token = None
        self.__expires = None

        # Setup the Session Details based on if Old or New API.
        codeduser = quote_plus(username)
        if self.__api_version == "v1":
            hashpass = sha1(password.encode("utf-8")).hexdigest()
            self.__login_params = (
                f"login=login&uname={codeduser}&upwd={hashpass}&{APP_CLIENTINFO}"
            )
        else:
            codedpass = quote_plus(password)
            self.__login_params = (
                f"grant_type=password&username={codeduser}&"
                + f"password={codedpass}&{APP_CLIENTINFO_NEW}"
            )

    async def __token_expired(self) -> bool:
        """Return if the token has expired."""
        if self.__expires is None:
            return True

        if self.__api_version == "v1":
            if self.__expires <= datetime.fromtimestamp(time.mktime(time.gmtime())):
                return True
        else:
            if self.__expires <= datetime.now():
                return True

        return False

    async def __post(self, url: str, params: str) -> dict:
        """Push updates to the API."""
        if await self.__token_expired():
            _LOGGER.info("__post: Token Expired, Refreshing Token.")
            await self.__connect_refresh()

        _LOGGER.debug("__post: data to: %s, params: %s", url, params)
        try:
            if self.__api_version == "v1":
                # Original uses post, with Cookie Token
                response = await self.__session.post(
                    urljoin(URL_BASE, url),
                    data=params,
                    headers={"Content-Type": "application/x-www-form-urlencoded"},
                    cookies={"PHPSESSID": self.__token, "$version": "1"},
                )
            else:
                # New uses get, with Authorization Bearer
                response = await self.__session.post(
                    urljoin(URL_BASE_NEW, url),
                    data=params,
                    headers={
                        "Authorization": f"Bearer {self.__token}",
                        "Content-Type": "application/x-www-form-urlencoded",
                        "Host": "mastertherm.online",
                        "Connection": "close",
                    },
                )

            _LOGGER.debug(
                "__post: response status: %s, content: %s",
                response.status,
                await response.text(),
            )
            response_json = await response.json()
        except ClientConnectionError as ex:
            _LOGGER.error("Client Connection Error: %s", ex)
            raise MasterthermConnectionError("3", "Client Connection Error") from ex
        except JSONDecodeError as ex:
            response_text = await response.text()
            _LOGGER.error(
                "JSON Decode Error: %s:%s",
                response.status,
                response_text,
            )
            raise MasterthermResponseFormatError(response.status, response_text) from ex
        except ContentTypeError as ex:
            if response.status == 504:
                raise MasterthermServerTimeoutError(
                    response.status, response.reason
                ) from ex

            response_text = await response.text()
            if response_text == "User not logged in":
                raise MasterthermTokenInvalid(response.status, response_text) from ex
            else:
                _LOGGER.error(
                    "Mastertherm API some other error: %s:%s",
                    response.status,
                    response_text,
                )
                raise MasterthermConnectionError(response.status, response_text) from ex

        # Version 2 responds with an error and json.
        # We should only get something other than 200 if the servers are down.
        if response.status != 200:
            # Deal with the v2 error if we have a response
            if response_json["status"]["id"] == 401:
                raise MasterthermTokenInvalid(response.status, response_json)
            else:
                _LOGGER.error("Mastertherm API some other error: %s", response_json)
                raise MasterthermResponseFormatError(response.status, response_json)

        return response_json

    async def __get(self, url: str, params: str) -> dict:
        """Get updates from the API, for old this mostly uses Post."""
        if await self.__token_expired():
            _LOGGER.info("__get: Token Expired, Refreshing Token.")
            await self.__connect_refresh()

        _LOGGER.debug("__get: data from: %s", url)
        try:
            if self.__api_version == "v1":
                # Original uses post, with Cookie Token
                response = await self.__session.post(
                    urljoin(URL_BASE, url),
                    data=params,
                    headers={"content-type": "application/x-www-form-urlencoded"},
                    cookies={"PHPSESSID": self.__token, "$version": "1"},
                )
            else:
                # New uses get, with Authorization Bearer
                response = await self.__session.get(
                    urljoin(URL_BASE_NEW, url),
                    params=params,
                    headers={
                        "Authorization": f"Bearer {self.__token}",
                        "Host": "mastertherm.online",
                        "Connection": "close",
                    },
                )

            _LOGGER.debug(
                "__get: response status: %s, content: %s",
                response.status,
                await response.text(),
            )
            response_json = await response.json()
        except ClientConnectionError as ex:
            _LOGGER.error("Client Connection Error: %s", ex)
            raise MasterthermConnectionError("3", "Client Connection Error") from ex
        except JSONDecodeError as ex:
            response_text = await response.text()
            _LOGGER.error(
                "JSON Decode Error: %s:%s",
                response.status,
                response_text,
            )
            raise MasterthermResponseFormatError(response.status, response_text) from ex
        except ContentTypeError as ex:
            if response.status == 504:
                raise MasterthermServerTimeoutError(
                    response.status, response.reason
                ) from ex

            response_text = await response.text()
            if response_text == "User not logged in":
                raise MasterthermTokenInvalid(response.status, response_text) from ex
            else:
                _LOGGER.error(
                    "Mastertherm API some other error: %s:%s",
                    response.status,
                    response_text,
                )
                raise MasterthermConnectionError(response.status, response_text) from ex

        # Version 2 responds with an error and json.
        if response.status != 200:
            # Deal with the v2 error if we have a response
            if response_json["status"]["id"] == 401:
                raise MasterthermTokenInvalid(response.status, response_json)
            else:
                _LOGGER.error("Mastertherm API some other error: %s", response_json)
                raise MasterthermResponseFormatError(response.status, response_json)

        return response_json

    async def __connect_refresh(self) -> dict:
        """Perform the Connect only or refresh token."""
        # Connect based on requirements
        if self.__api_version == "v1":
            # Clear out cookies, clears the auth token.
            url = urljoin(URL_BASE, URL_LOGIN)
        else:
            url = urljoin(URL_BASE_NEW, URL_LOGIN_NEW)

        try:
            response = await self.__session.post(
                url,
                data=self.__login_params,
                cookies={"PHPSESSID": "", "$version": "1"},
                headers={"content-type": "application/x-www-form-urlencoded"},
            )
        except ClientConnectionError as ex:
            _LOGGER.warning("Connection Error: %s", ex)
            raise MasterthermConnectionError("3", "Client Connection Error") from ex

        # Response shoudl always be 200 even for login failures
        if response.status != 200:
            error_msg = await response.text()
            _LOGGER.warning(
                "Site Error status=%s, Error=%s", response.status, error_msg
            )
            raise MasterthermConnectionError(str(response.status), error_msg)

        response_json = await response.json()

        # Separate process for Old and New API's
        if self.__api_version == "v1":
            # Old Process gets return and modules in one go
            # token and expiry is stored in a cookie.
            if response_json["returncode"] != 0:
                raise MasterthermAuthenticationError(
                    response_json["returncode"], response_json["message"]
                )

            # Get or Refresh the Token and Expiry
            self.__token = response.cookies["PHPSESSID"].value
            self.__expires = datetime.strptime(
                response.cookies["PHPSESSID"]["expires"], DATE_FORMAT
            )
        else:
            # New process uses JSON responses with Openconnect ID.
            # Requires an additional call to get the modules.
            if "error" in response_json:
                raise MasterthermAuthenticationError(
                    response_json["error"], response_json["error_description"]
                )

            #  Get or Refresh the Token and Expiry
            self.__token = response_json["access_token"]
            self.__expires = datetime.today() + timedelta(
                seconds=response_json["expires_in"]
            )

        return response_json

    def get_url(self) -> str:
        """Return the API URL Used.

        Returns:
            URL(str): The API URL for the version.

        """
        if self.__api_version == "v1":
            return URL_BASE
        else:
            return URL_BASE_NEW

    async def connect(self) -> dict:
        """Perform the connection to the Mastertherm API Server.

        Returns:
             devices (dict): Return the list of devices, modules and units.

        Raises:
            MasterthermConnectionError - Failed to Connect
            MasterthermAuthenticationError - Failed to Authenticate
            MasterthermUnsupportedRole - Role is not in supported roles

        """
        response_json = await self.__connect_refresh()
        if self.__api_version == "v2":
            # Get the Modules as this now has moved to outside of the auth process
            response_json = await self.__get(url=URL_MODULES_NEW, params=None)

        # Next is same for old and new process, doing a double check just incase
        if response_json["returncode"] != 0:
            raise MasterthermAuthenticationError(
                response_json["returncode"], response_json["message"]
            )

        # Check if role is supported
        if response_json["role"] not in SUPPORTED_ROLES:
            raise MasterthermUnsupportedRole(
                "2", "Unsupported Role " + response_json["role"]
            )

        return response_json

    async def get_device_info(self, module_id: str, unit_id: str) -> dict:
        """Get the Device information.

        Args:
            module_id: This is the module_id for the unit
            unit_id: This is the unit id for the unit

        Return:
            device_info (dict): Information for a specific device.

        Raises:
            MasterthermConnectionError - General Connection Issue
            MasterthermTokenInvalid - Token has expired or is invalid
            MasterthermResponseFormatError - Some other issue, probably temporary
            MasterthermServerTimeoutError - Server Timed Out more than once.

        """
        retry = False
        params = f"moduleid={module_id}&unitid={unit_id}&application=android"

        _LOGGER.info("Get Device Info %s:%s", module_id, unit_id)
        try:
            response_json = await self.__get(
                url=URL_PUMPINFO if self.__api_version == "v1" else URL_PUMPINFO_NEW,
                params=params,
            )
        except MasterthermTokenInvalid as ex:
            _LOGGER.info("Token Expired Early Retry: %s:%s", ex.status, ex.message)
            retry = True
            self.__expires = None
        except MasterthermServerTimeoutError as ex:
            _LOGGER.info("API Timed Out Retry: %s:%s", ex.status, ex.message)
            retry = True

        if retry:
            await asyncio.sleep(0.5)
            response_json = await self.__get(
                url=URL_PUMPINFO if self.__api_version == "v1" else URL_PUMPINFO_NEW,
                params=params,
            )

        return response_json

    async def get_device_data(
        self, module_id: str, unit_id: str, last_update_time: str = None
    ) -> dict:
        """Get the Device lastest data.

        Args:
            module_id: This is the module_id for the unit
            unit_id: This is the unit id for the unit
            last_update_time: Optional last update date in number format

        Return:
            device_data (dict): data or updated data for a specific device.

        Raises:
            MasterthermConnectionError - General Connection Issue
            MasterthermTokenInvalid - Token has expired or is invalid
            MasterthermResponseFormatError - Some other issue, probably temporary
            MasterthermPumpDisconnected - Pump is unavailable, disconnected or offline.
            MasterthermServerTimeoutError - Server Timed Out more than once.

        """
        retry = False
        params = f"moduleId={module_id}&deviceId={unit_id}&application=android&"
        if last_update_time is None:
            params = (
                params
                + "messageId=1&lastUpdateTime=0&errorResponse=true&fullRange=true"
            )
        else:
            params = (
                params
                + f"messageId=2&lastUpdateTime={last_update_time}&errorResponse=true&fullRange=true"
            )

        _LOGGER.info("Get Device Data %s:%s", module_id, unit_id)
        try:
            response_json = await self.__get(
                url=URL_PUMPDATA if self.__api_version == "v1" else URL_PUMPDATA_NEW,
                params=params,
            )
        except MasterthermTokenInvalid as ex:
            _LOGGER.info("Token Expired Early Retrying: %s:%s", ex.status, ex.message)
            retry = True
            self.__expires = None
        except MasterthermServerTimeoutError as ex:
            _LOGGER.info("API Timed Out Retry: %s:%s", ex.status, ex.message)
            retry = True

        if retry:
            await asyncio.sleep(0.5)
            response_json = await self.__get(
                url=URL_PUMPDATA if self.__api_version == "v1" else URL_PUMPDATA_NEW,
                params=params,
            )

        # Check for Errors with the Pump.
        error_id = response_json["error"]["errorId"]
        if error_id != 0:
            raise MasterthermPumpError(error_id, response_json["error"]["errorMessage"])

        # No error process the response.
        if response_json["data"] != {}:
            data_file = ""
            if self.__api_version == "v1":
                if "varfile_mt1_config1" in response_json["data"]:
                    data_file = "varfile_mt1_config1"
                elif "varfile_mt1_config2" in response_json["data"]:
                    data_file = "varfile_mt1_config2"
            elif self.__api_version == "v2":
                data_file = "varFileData"

            response_json["data"]["varData"] = response_json["data"][data_file]
            del response_json["data"][data_file]

            # Sort the Data
            sorted_reg = {}
            device_reg = response_json["data"]["varData"][str(unit_id).zfill(3)]
            for key in natsorted(device_reg.keys()):
                sorted_reg[key] = device_reg[key]
            response_json["data"]["varData"][str(unit_id).zfill(3)] = sorted_reg

        return response_json

    async def set_device_data(
        self, module_id: str, unit_id: str, register: str, value: any
    ) -> bool:
        """Set device data a specific register to a specific value.

        Updating any registry setting can cause the system to stop working
        the controller only allows tested updates this API has no protection.

        Args:
            module_id: This is the module_id for the unit
            unit_id: This is the unit id for the unit
            register: The Register to update
            value: The value to set.

        Return:
           success (bool): return true if succes and false if not.

        Raises:
            MasterthermConnectionError - General Connection Issue
            MasterthermTokenInvalid - Token has expired or is invalid
            MasterthermResponseFormatError - Some other issue, probably temporary
            MasterthermServerTimeoutError - Server Timed Out more than once.

        """
        retry = False
        params = (
            f"moduleId={module_id}&deviceId={unit_id}&"
            + "configFile=varfile_mt1_config&messageId=1&errorResponse=true&"
            + f"variableId={register}&variableValue={value}&application=android"
        )

        _LOGGER.info("Set Device Reg %s:%s:%s:%s", module_id, unit_id, register, value)
        try:
            response_json = await self.__post(
                url=(
                    URL_POSTUPDATE if self.__api_version == "v1" else URL_POSTUPDATE_NEW
                ),
                params=params,
            )
        except MasterthermTokenInvalid as ex:
            _LOGGER.info("Token Expired Early Retrying: %s:%s", ex.status, ex.message)
            retry = True
            self.__expires = None
        except MasterthermServerTimeoutError as ex:
            _LOGGER.info("API Timed Out Retry: %s:%s", ex.status, ex.message)
            retry = True

        if retry:
            await asyncio.sleep(0.5)
            response_json = await self.__post(
                url=(
                    URL_POSTUPDATE if self.__api_version == "v1" else URL_POSTUPDATE_NEW
                ),
                params=params,
            )

        _LOGGER.info("Set Device Reg response: %s", response_json)

        # Check the response for any errors:
        if response_json["error"]["errorId"] != 0:
            _LOGGER.error(
                "Client Connection Error: %s, Message: %s",
                response_json["error"]["errorId"],
                response_json["error"]["errorMessage"],
            )
            return False

        # Get the Value that is returned and set.
        if self.__api_version == "v1":
            data_file = ""
            if "varfile_mt1_config1" in response_json["data"]:
                data_file = "varfile_mt1_config1"
            elif "varfile_mt1_config2" in response_json["data"]:
                data_file = "varfile_mt1_config2"

            set_value = response_json["data"][data_file][str(unit_id).zfill(3)][
                register
            ]
        else:
            set_value = response_json["data"]["data"][str(unit_id).zfill(3)][register]

        return set_value == value



"""Test the API Client."""
import json

from datetime import datetime, timedelta
from hashlib import sha1
from unittest.mock import patch

from aiohttp import web, ServerTimeoutError
from aiohttp.test_utils import AioHTTPTestCase, Request

import pytest

from masterthermconnect import (
    MasterthermAuthenticationError,
    MasterthermConnectionError,
    MasterthermTokenInvalid,
    MasterthermPumpError,
    MasterthermUnsupportedVersion,
    MasterthermResponseFormatError,
)
from masterthermconnect.api import MasterthermAPI
from masterthermconnect.const import (
    DATE_FORMAT,
    URL_LOGIN,
    URL_PUMPDATA,
    URL_PUMPINFO,
    URL_POSTUPDATE,
)

from .conftest import GENERAL_ERROR_RESPONSE, VALID_LOGIN, load_fixture


@patch("masterthermconnect.api.URL_BASE", "")
class APITestCase(AioHTTPTestCase):
    """Test the Original API Connection."""

    error_type = ""
    data: dict = None
    info: dict = None

    async def get_application(self):  # noqa: C901
        """Start and Return a mock application."""

        async def _connect_response(request: Request):
            """Check the Test Login Credentials and return login connect or failure."""
            data = await request.post()
            password = sha1(VALID_LOGIN["upwd"].encode("utf-8")).hexdigest()

            if self.error_type == "ClientConnectionError":
                return None

            if data["uname"] == VALID_LOGIN["uname"] and data["upwd"] == password:
                response_text = load_fixture("login_success.json")
            else:
                response_text = load_fixture("login_invalid.json")

            if self.error_type == "token_expire":
                token_expires = datetime.now() + timedelta(seconds=0)
            else:
                token_expires = datetime.now() + timedelta(seconds=60)

            response = web.Response(
                text=response_text,
                content_type="application/json",
            )
            response.set_cookie(
                "PHPSESSID",
                VALID_LOGIN["token"],
                expires=token_expires.strftime(DATE_FORMAT) + "GMT",
            )

            return response

        async def _get_info(request: Request):
            """Get the Pump Info."""
            data = await request.post()
            content_type = "application/json"

            if self.error_type == "ClientConnectionError":
                return None

            if self.error_type.startswith("login"):
                response_text = GENERAL_ERROR_RESPONSE
                content_type = "text/plain"
                if self.error_type == "login_once":
                    self.error_type = ""
            elif self.error_type.startswith("timeout"):
                if self.error_type == "timeout_once":
                    self.error_type = ""
                    raise ServerTimeoutError("Connection timeout.")
            elif self.info is None:
                module_id = data["moduleid"]
                unit_id = data["unitid"]
                response_text = load_fixture(f"pumpinfo_{module_id}_{unit_id}.json")
                if response_text is not None:
                    self.info = json.loads(response_text)
            else:
                response_text = json.dumps(self.info)

            if response_text is None:
                response_text = load_fixture("pumpinfo_invalid.json")

            response = web.Response(text=response_text, content_type=content_type)
            return response

        async def _get_data(request: Request):
            """Get the Pump Data."""
            data = await request.post()
            content_type = "application/json"
            module_id = data["moduleId"]
            unit_id = data["deviceId"]
            last_update_time = data["lastUpdateTime"]

            if self.error_type == "ClientConnectionError":
                return None

            if self.error_type == "content_error":
                response_text = "content error"
                content_type = "text/plain"
            elif self.error_type == "json_error":
                response_text = "json error"
            elif self.error_type.startswith("login"):
                response_text = GENERAL_ERROR_RESPONSE
                content_type = "text/plain"
                if self.error_type == "login_once":
                    self.error_type = ""
            elif self.error_type.startswith("timeout"):
                if self.error_type == "timeout_once":
                    self.error_type = ""
                    raise ServerTimeoutError("Connection timeout.")
            elif self.error_type == "unavailable":
                response_text = load_fixture("pumpdata_unavailable.json")
            elif self.data is None or last_update_time != "0":
                response_text = load_fixture(
                    f"pumpdata_{module_id}_{unit_id}_{last_update_time}.json"
                )
            else:
                response_text = json.dumps(self.data)

            if response_text is None:
                response_text = load_fixture("pumpdata_invalid.json")
            elif content_type == "application/json" and self.error_type != "json_error":
                self.data = json.loads(response_text)

            response = web.Response(text=response_text, content_type=content_type)
            return response

        async def _set_data(request: Request):
            """Set the Pump Data."""
            data = await request.post()
            content_type = "application/json"
            # module_id = data["moduleId"]
            unit_id = str(data["deviceId"]).zfill(3)
            variable_id = data["variableId"]
            variable_value = data["variableValue"]

            if self.error_type.startswith("login"):
                response_text = GENERAL_ERROR_RESPONSE
                content_type = "text/plain"
                if self.error_type == "login_once":
                    self.error_type = ""
            elif self.error_type.startswith("timeout"):
                if self.error_type == "timeout_once":
                    self.error_type = ""
                    raise ServerTimeoutError("Connection timeout.")
            elif self.error_type == "write_error":
                response_text = load_fixture("pumpwrite_error.json")
                self.error_type = ""
            else:
                response_text = load_fixture("pumpwrite_success.json")
                if "varfile_mt1_config1" in self.data["data"]:
                    self.data["data"]["varfile_mt1_config1"][unit_id][
                        variable_id
                    ] = variable_value
                else:
                    self.data["data"]["varfile_mt1_config2"][unit_id][
                        variable_id
                    ] = variable_value

            response = web.Response(text=response_text, content_type=content_type)
            return response

        app = web.Application()
        app.router.add_post(URL_LOGIN, _connect_response)
        app.router.add_post(URL_PUMPINFO, _get_info)
        app.router.add_post(URL_PUMPDATA, _get_data)
        app.router.add_post(URL_POSTUPDATE, _set_data)
        return app

    async def test_setup(self):
        """Test the API Setup."""
        assert MasterthermAPI(
            VALID_LOGIN["uname"], VALID_LOGIN["upwd"], self.client, api_version="v1"
        )

    async def test_connect(self):
        """Test the API Logs in and Returns modules."""
        api = MasterthermAPI(
            VALID_LOGIN["uname"], VALID_LOGIN["upwd"], self.client, api_version="v1"
        )
        result = await api.connect()

        assert result != {}
        assert result["returncode"] == 0
        assert result["modules"][0]["id"] == "1234"
        assert result["role"] == "400"

    async def test_autherror(self):
        """Test a Connection Authentication Error."""
        api = MasterthermAPI(
            VALID_LOGIN["uname"],
            VALID_LOGIN["upwd"] + "bad",
            self.client,
            api_version="v1",
        )

        with pytest.raises(MasterthermAuthenticationError):
            await api.connect()

    async def test_connect_unavailabe(self):
        """Test the Connection unavailable or invalid."""
        api = MasterthermAPI(
            VALID_LOGIN["uname"], VALID_LOGIN["upwd"], self.client, api_version="v1"
        )

        with patch("masterthermconnect.api.URL_LOGIN", "/"), pytest.raises(
            MasterthermConnectionError
        ):
            await api.connect()

    async def test_getinfo(self):
        """Test returning the Device Information."""
        api = MasterthermAPI(
            VALID_LOGIN["uname"], VALID_LOGIN["upwd"], self.client, api_version="v1"
        )
        assert await api.connect() != {}

        info = await api.get_device_info("1234", "1")

        assert info != {}
        assert info["moduleid"] == "1234"
        assert info["type"] == "AQI"

    async def test_getinfo_tokeninvalid(self):
        """Test the get device info token is not valid."""
        api = MasterthermAPI(
            VALID_LOGIN["uname"], VALID_LOGIN["upwd"], self.client, api_version="v1"
        )
        assert await api.connect() is not {}

        self.error_type = "login"
        with pytest.raises(MasterthermTokenInvalid):
            await api.get_device_info("1234", "1")

    async def test_getinfo_invalid(self):
        """Test the device info for invalid device."""
        api = MasterthermAPI(
            VALID_LOGIN["uname"], VALID_LOGIN["upwd"], self.client, api_version="v1"
        )
        assert await api.connect() is not {}
        info = await api.get_device_info("1234", "2")

        assert info["returncode"] != 0

    async def test_getdata(self):
        """Test the Get Device data from new."""
        api = MasterthermAPI(
            VALID_LOGIN["uname"], VALID_LOGIN["upwd"], self.client, api_version="v1"
        )
        assert await api.connect() is not {}

        data = await api.get_device_data("1234", "1")

        assert data != {}
        assert data["error"]["errorId"] == 0
        assert data["messageId"] == 1
        assert data["data"] != {}

    async def test_getdata_update(self):
        """Test the Get device Data update."""
        api = MasterthermAPI(
            VALID_LOGIN["uname"], VALID_LOGIN["upwd"], self.client, api_version="v1"
        )
        assert await api.connect() is not {}

        data = await api.get_device_data("1234", "1")
        assert data != {}

        last_update_time = data["timestamp"]
        a_500 = data["data"]["varData"]["001"]["A_500"]

        data = await api.get_device_data("1234", "1", last_update_time=last_update_time)
        assert data != {}
        assert data["error"]["errorId"] == 0
        assert data["timestamp"] != last_update_time
        assert data["data"]["varData"]["001"]["A_500"] != a_500

    async def test_getdata_invalid(self):
        """Test the get invalid device."""
        api = MasterthermAPI(
            VALID_LOGIN["uname"], VALID_LOGIN["upwd"], self.client, api_version="v1"
        )
        assert await api.connect() is not {}

        try:
            await api.get_device_data("1234", "2")
        except MasterthermPumpError as ex:
            assert ex.status == MasterthermPumpError.DEVICENOTFOUND

    async def test_offline_initial(self):
        """Test getting data when the HP is offline on initial connect."""
        api = MasterthermAPI(
            VALID_LOGIN["uname"], VALID_LOGIN["upwd"], self.client, api_version="v1"
        )
        assert await api.connect() is not {}

        self.error_type = "unavailable"
        try:
            await api.get_device_data("1234", "1")
        except MasterthermPumpError as ex:
            assert ex.status == MasterthermPumpError.OFFLINE

    async def test_offline_update(self):
        """Test getting data when the HP is offline on update."""
        api = MasterthermAPI(
            VALID_LOGIN["uname"], VALID_LOGIN["upwd"], self.client, api_version="v1"
        )
        assert await api.connect() is not {}

        data = await api.get_device_data("1234", "1")
        assert data != {}

        self.error_type = "unavailable"
        try:
            await api.get_device_data("1234", "1", last_update_time=None)
        except MasterthermPumpError as ex:
            assert ex.status == MasterthermPumpError.OFFLINE

    async def test_setdata(self):
        """Test we can send data to update the API."""
        api = MasterthermAPI(
            VALID_LOGIN["uname"], VALID_LOGIN["upwd"], self.client, api_version="v1"
        )
        assert await api.connect() is not {}

        # Read the Data and issue, Power State
        data = await api.get_device_data("1234", "1")
        assert data["data"]["varData"]["001"]["D_3"] == "1"

        # Issue an update back to the API
        assert await api.set_device_data("1234", "1", "D_3", "0")

        # Re-Read to check the update is correct.
        data = await api.get_device_data("1234", "1")
        assert data["data"]["varData"]["001"]["D_3"] == "0"

    async def test_setdata_fail(self):
        """Test we can send data to update the API."""
        api = MasterthermAPI(
            VALID_LOGIN["uname"], VALID_LOGIN["upwd"], self.client, api_version="v1"
        )
        assert await api.connect() is not {}

        # Read the Data and issue, Power State
        data = await api.get_device_data("1234", "1")
        assert data["data"]["varData"]["001"]["D_3"] == "1"

        # Issue an update back to the API
        self.error_type = "write_error"
        assert await api.set_device_data("1234", "1", "D_3", "0") is False

        # Re-Read to check the update is correct.
        data = await api.get_device_data("1234", "1")
        assert data["data"]["varData"]["001"]["D_3"] == "1"

    async def test_unsupported_version(self):
        """Test exception for unsupported version."""
        with pytest.raises(MasterthermUnsupportedVersion):
            MasterthermAPI("1234", "1", self.client, api_version="v10")

    async def test_expired_token(self):
        """Test for an expired token and re-connect."""
        api = MasterthermAPI(
            VALID_LOGIN["uname"], VALID_LOGIN["upwd"], self.client, api_version="v1"
        )
        self.error_type = "token_expire"
        assert await api.connect() is not {}

        assert await api.get_device_info("1234", "1")
        assert await api.get_device_data("1234", "1")
        assert await api.set_device_data("1234", "1", "D_3", "0")

    async def test_invalid_token(self):
        """Test for an invalid token and re-try."""
        api = MasterthermAPI(
            VALID_LOGIN["uname"], VALID_LOGIN["upwd"], self.client, api_version="v1"
        )
        assert await api.connect() is not {}

        assert await api.get_device_info("1234", "1")

    async def test_client_connection_error(self):
        """Test for Client Connection Errors, such as timeout."""
        api = MasterthermAPI(
            VALID_LOGIN["uname"], VALID_LOGIN["upwd"], self.client, api_version="v1"
        )
        assert await api.connect()
        assert await api.get_device_info("1234", "1")

        self.error_type = "ClientConnectionError"
        with pytest.raises(MasterthermConnectionError):
            await api.connect()

        with pytest.raises(MasterthermConnectionError):
            await api.get_device_info("1234", "1")

        with pytest.raises(MasterthermConnectionError):
            await api.set_device_data("1234", "1", "D_3", "0")

    async def test_response_errors(self):
        """Test response format and content errors."""
        api = MasterthermAPI(
            VALID_LOGIN["uname"], VALID_LOGIN["upwd"], self.client, api_version="v1"
        )
        assert await api.connect()
        assert await api.get_device_info("1234", "1")

        self.error_type = "content_error"
        with pytest.raises(MasterthermConnectionError):
            await api.get_device_data("1234", "1")

        self.error_type = "json_error"
        with pytest.raises(MasterthermResponseFormatError):
            await api.get_device_data("1234", "1")

    async def test_fail_with_retry(self):
        """Test token error with a retry."""
        api = MasterthermAPI(
            VALID_LOGIN["uname"], VALID_LOGIN["upwd"], self.client, api_version="v1"
        )
        assert await api.connect()

        self.error_type = "login_once"
        assert await api.get_device_info("1234", "1")

        self.error_type = "login_once"
        assert await api.get_device_data("1234", "1")

        self.error_type = "login_once"
        await api.set_device_data("1234", "1", "D_3", "0")

    async def test_timeout(self):
        """Test timeout re-tries on refresh."""
        api = MasterthermAPI(
            VALID_LOGIN["uname"], VALID_LOGIN["upwd"], self.client, api_version="v1"
        )
        assert await api.connect()

        self.error_type = "timeout_once"
        assert await api.get_device_info("1234", "1")

        self.error_type = "timeout_once"
        assert await api.get_device_data("1234", "1")

        self.error_type = "timeout_once"
        await api.set_device_data("1234", "1", "D_3", "0")



"""Test the Main command."""
import json

from unittest.mock import patch

from masterthermconnect.__main__ import main as MasterthermConnect

from .conftest import VALID_LOGIN, ConnectionMock


def test_help(capsys):
    """Test the Main Help."""
    MasterthermConnect(["--help"])

    out, err = capsys.readouterr()
    assert out.startswith(
        "usage: masterthermconnect [-h] [--version] [-d] [-v] {get,set}"
    )
    assert err == ""


def test_get_help(capsys):
    """Test the Get Help."""
    MasterthermConnect(["get", "-h"])

    out, err = capsys.readouterr()
    assert out.startswith(
        "usage: masterthermconnect get [-h] [--user USER] [--password PASSWORD]"
    )
    assert err == ""


def test_set_help(capsys):
    """Test the Set Help."""
    MasterthermConnect(["set", "-h"])

    out, err = capsys.readouterr()
    assert out.startswith(
        "usage: masterthermconnect set [-h] [--user USER] [--password PASSWORD]"
    )
    assert err == ""


def test_get_v1(capsys):
    """Test the Get devices for v1."""
    mockconnect = ConnectionMock(api_version="v1")

    with patch(
        "masterthermconnect.api.MasterthermAPI.connect",
        return_value=mockconnect.connect(),
    ) as mock_api_connect, patch(
        "masterthermconnect.api.MasterthermAPI.get_device_info",
        side_effect=mockconnect.get_device_info,
    ) as mock_get_device_info, patch(
        "masterthermconnect.api.MasterthermAPI.get_device_data",
        side_effect=mockconnect.get_device_data,
    ) as mock_get_device_data:
        return_code = MasterthermConnect(
            [
                "get",
                "--user",
                VALID_LOGIN["uname"],
                "--password",
                VALID_LOGIN["upwd"],
                "devices",
            ]
        )

        out, err = capsys.readouterr()

    assert len(mock_api_connect.mock_calls) > 0
    assert len(mock_get_device_info.mock_calls) > 0
    assert len(mock_get_device_data.mock_calls) > 0

    assert return_code == 0
    assert err == ""

    devices = json.loads(out.replace("\n", ""))
    assert devices["1234_1"]["module_id"] == "1234"


def test_set_v1(capsys):
    """Test the Get devices for v1."""
    mockconnect = ConnectionMock(api_version="v1")

    with patch(
        "masterthermconnect.api.MasterthermAPI.connect",
        return_value=mockconnect.connect(),
    ), patch(
        "masterthermconnect.api.MasterthermAPI.get_device_info",
        side_effect=mockconnect.get_device_info,
    ), patch(
        "masterthermconnect.api.MasterthermAPI.get_device_data",
        side_effect=mockconnect.get_device_data,
    ), patch(
        "masterthermconnect.api.MasterthermAPI.set_device_data",
        side_effect=mockconnect.set_device_data,
    ):
        return_code = MasterthermConnect(
            [
                "set",
                "--user",
                VALID_LOGIN["uname"],
                "--password",
                VALID_LOGIN["upwd"],
                "data",
                "1234",
                "1",
                "heating_circuits.hc1.ambient_requested",
                "50.1",
            ]
        )

        out, err = capsys.readouterr()

    assert return_code == 0
    assert err == ""
    assert out == "Data after Update: heating_circuits.hc1.ambient_requested = 50.1\n"


def test_set_power(capsys):
    """Test the Set devices for power state."""
    mockconnect = ConnectionMock(api_version="v1")

    with patch(
        "masterthermconnect.api.MasterthermAPI.connect",
        return_value=mockconnect.connect(),
    ), patch(
        "masterthermconnect.api.MasterthermAPI.get_device_info",
        side_effect=mockconnect.get_device_info,
    ), patch(
        "masterthermconnect.api.MasterthermAPI.get_device_data",
        side_effect=mockconnect.get_device_data,
    ), patch(
        "masterthermconnect.api.MasterthermAPI.set_device_data",
        side_effect=mockconnect.set_device_data,
    ):
        return_code = MasterthermConnect(
            [
                "set",
                "--user",
                VALID_LOGIN["uname"],
                "--password",
                VALID_LOGIN["upwd"],
                "data",
                "1234",
                "1",
                "hp_power_state",
                "false",
            ]
        )

        out, err = capsys.readouterr()

    assert return_code == 0
    assert err == ""
    assert out == "Data after Update: hp_power_state = False\n"


def test_set_reg(capsys):
    """Test the Set device registry."""
    mockconnect = ConnectionMock(api_version="v1")

    with patch(
        "masterthermconnect.api.MasterthermAPI.connect",
        return_value=mockconnect.connect(),
    ), patch(
        "masterthermconnect.api.MasterthermAPI.get_device_info",
        side_effect=mockconnect.get_device_info,
    ), patch(
        "masterthermconnect.api.MasterthermAPI.get_device_data",
        side_effect=mockconnect.get_device_data,
    ), patch(
        "masterthermconnect.api.MasterthermAPI.set_device_data",
        side_effect=mockconnect.set_device_data,
    ):
        return_code = MasterthermConnect(
            [
                "set",
                "--user",
                VALID_LOGIN["uname"],
                "--password",
                VALID_LOGIN["upwd"],
                "reg",
                "--force",
                "1234",
                "1",
                "D_3",
                "1",
            ]
        )

        out, err = capsys.readouterr()

    assert return_code == 0
    assert err == ""
    assert out == "Registration after Update: D_3 = 1\n"



"""Test the New API Client."""
import json

from unittest.mock import patch

from aiohttp import web
from aiohttp.test_utils import AioHTTPTestCase, Request

import pytest

from masterthermconnect import (
    MasterthermAuthenticationError,
    MasterthermConnectionError,
    MasterthermResponseFormatError,
    MasterthermTokenInvalid,
    MasterthermPumpError,
    MasterthermUnsupportedVersion,
)
from masterthermconnect.api import MasterthermAPI
from masterthermconnect.const import (
    URL_LOGIN_NEW,
    URL_MODULES_NEW,
    URL_PUMPDATA_NEW,
    URL_PUMPINFO_NEW,
    URL_POSTUPDATE_NEW,
)

from .conftest import VALID_LOGIN, load_fixture


@patch("masterthermconnect.api.URL_BASE_NEW", "")
class APINewTestCase(AioHTTPTestCase):
    """Test the New API Connection."""

    error_type = ""
    data: dict = None
    info: dict = None

    async def get_application(self):
        """Start and Return a mock application."""

        async def _connect_response(request: Request):
            """Check the Test Login Credentials and return login connect or failure."""
            data = await request.post()

            if self.error_type == "ClientConnectionError":
                return None

            if (
                data["username"] == VALID_LOGIN["uname"]
                and data["password"] == VALID_LOGIN["upwd"]
            ):
                response_text = load_fixture("newapi/login_success.json")

                if self.error_type == "token_expire":
                    response_text = response_text.replace(": 7200", ": 0")
                else:
                    self.data = None
                    self.info = None
            else:
                response_text = load_fixture("newapi/login_invalid.json")

            response = web.Response(
                text=response_text,
                content_type="application/json",
            )

            return response

        async def _get_modules(request: Request):
            """Get the Devices, new in the new API."""
            token = request.headers.get("Authorization")
            content_type = "application/json"
            response_status = 200
            if self.error_type == "login" or token != "Bearer bearertoken":
                response_text = load_fixture("newapi/error_authorization.json")
                response_status = 401
            else:
                response_text = load_fixture("newapi/modules.json")

            if response_text is None:
                response_text = load_fixture("newapi/error_unavailable.json")
                response_status = 500

            response = web.Response(
                status=response_status, text=response_text, content_type=content_type
            )
            return response

        async def _get_info(request: Request):
            """Get the Pump Info."""
            token = request.headers.get("Authorization")
            data = request.query
            content_type = "application/json"
            response_status = 200

            if self.error_type == "ClientConnectionError":
                return None

            if self.error_type == "login" or token != "Bearer bearertoken":
                response_text = load_fixture("newapi/error_authorization.json")
                response_status = 401
            elif self.info is None:
                module_id = data["moduleid"]
                unit_id = data["unitid"]
                response_text = load_fixture(
                    f"newapi/pumpinfo_{module_id}_{unit_id}.json"
                )
            else:
                response_text = json.dumps(self.info)

            if response_text is None:
                response_text = load_fixture("newapi/error_unavailable.json")
                response_status = 500
            else:
                self.info = json.loads(response_text)

            response = web.Response(
                status=response_status, text=response_text, content_type=content_type
            )
            return response

        async def _get_data(request: Request):
            """Get the Pump Info."""
            token = request.headers.get("Authorization")
            data = request.query

            if self.error_type == "ClientConnectionError":
                return None

            module_id = data["moduleId"]
            unit_id = data["deviceId"]
            last_update_time = data["lastUpdateTime"]

            content_type = "application/json"
            response_status = 200
            if self.error_type == "login" or token != "Bearer bearertoken":
                response_text = load_fixture("newapi/error_authorization.json")
                response_status = 401
            elif self.data is None or last_update_time != "0":
                response_text = load_fixture(
                    f"newapi/pumpdata_{module_id}_{unit_id}_{last_update_time}.json"
                )
            else:
                response_text = json.dumps(self.data)

            if response_text is None:
                response_text = load_fixture("newapi/error_unavailable.json")
                response_status = 500
            else:
                self.data = json.loads(response_text)

            response = web.Response(
                status=response_status, text=response_text, content_type=content_type
            )
            return response

        async def _set_data(request: Request):
            """Set the Pump Data."""
            data = await request.post()
            content_type = "application/json"
            # module_id = data["moduleId"]
            unit_id = str(data["deviceId"]).zfill(3)
            variable_id = data["variableId"]
            variable_value = data["variableValue"]

            response_text = load_fixture("newapi/pumpwrite_success.json")
            self.data["data"]["varFileData"][unit_id][variable_id] = variable_value

            response = web.Response(text=response_text, content_type=content_type)
            return response

        app: web.Application = web.Application()
        app.router.add_post(URL_LOGIN_NEW, _connect_response)
        app.router.add_get(URL_MODULES_NEW, _get_modules)
        app.router.add_get(URL_PUMPINFO_NEW, _get_info)
        app.router.add_get(URL_PUMPDATA_NEW, _get_data)
        app.router.add_post(URL_POSTUPDATE_NEW, _set_data)

        return app

    async def test_setup(self):
        """Test the API Setup."""
        assert MasterthermAPI(
            VALID_LOGIN["uname"], VALID_LOGIN["upwd"], self.client, api_version="v2"
        )

    async def test_connect(self):
        """Test the API Logs in and Returns modules."""
        api = MasterthermAPI(
            VALID_LOGIN["uname"], VALID_LOGIN["upwd"], self.client, api_version="v2"
        )
        result = await api.connect()

        assert result != {}
        assert result["returncode"] == 0
        assert result["modules"][0]["id"] == "10021"
        assert result["role"] == "400"

    async def test_autherror(self):
        """Test a Connection Authentication Error."""
        api = MasterthermAPI(
            VALID_LOGIN["uname"],
            VALID_LOGIN["upwd"] + "bad",
            self.client,
            api_version="v2",
        )

        with pytest.raises(MasterthermAuthenticationError):
            await api.connect()

    async def test_connecterror(self):
        """Test the Connection Invalid Error."""
        api = MasterthermAPI(
            VALID_LOGIN["uname"], VALID_LOGIN["upwd"], self.client, api_version="v2"
        )

        with patch("masterthermconnect.api.URL_LOGIN_NEW", "/"), pytest.raises(
            MasterthermConnectionError
        ):
            await api.connect()

    async def test_getinfo(self):
        """Test returning the Device Information."""
        api = MasterthermAPI(
            VALID_LOGIN["uname"], VALID_LOGIN["upwd"], self.client, api_version="v2"
        )
        assert await api.connect() != {}

        info = await api.get_device_info("10021", "1")

        assert info != {}
        assert info["moduleid"] == 10021
        assert info["type"] == "BAI"

    async def test_getinfo_tokeninvalid(self):
        """Test the get device info when token is invalid."""
        api = MasterthermAPI(
            VALID_LOGIN["uname"], VALID_LOGIN["upwd"], self.client, api_version="v2"
        )
        assert await api.connect() is not {}

        self.error_type = "login"
        with pytest.raises(MasterthermTokenInvalid):
            await api.get_device_info("10021", "1")

    async def test_getinfo_invalid(self):
        """Test the device info for invalid device."""
        api = MasterthermAPI(
            VALID_LOGIN["uname"], VALID_LOGIN["upwd"], self.client, api_version="v2"
        )
        assert await api.connect() is not {}

        with pytest.raises(MasterthermResponseFormatError):
            await api.get_device_info("1234", "2")

    async def test_getdata(self):
        """Test the Get Device data from new."""
        api = MasterthermAPI(
            VALID_LOGIN["uname"], VALID_LOGIN["upwd"], self.client, api_version="v2"
        )
        assert await api.connect() is not {}

        data = await api.get_device_data("10021", "1")

        assert data != {}
        assert data["error"]["errorId"] == 0
        assert data["messageId"] == 1
        assert data["data"] != {}

    async def test_getdata_update(self):
        """Test the Get device Data update."""
        api = MasterthermAPI(
            VALID_LOGIN["uname"], VALID_LOGIN["upwd"], self.client, api_version="v2"
        )
        assert await api.connect() is not {}

        data = await api.get_device_data("10021", "1")
        assert data != {}

        last_update_time = data["timestamp"]
        a_500 = data["data"]["varData"]["001"]["A_500"]

        data = await api.get_device_data(
            "10021", "1", last_update_time=last_update_time
        )
        assert data != {}
        assert data["error"]["errorId"] == 0
        assert data["timestamp"] != last_update_time
        assert data["data"]["varData"]["001"]["A_500"] != a_500

    async def test_getdata_invalid(self):
        """Test the get invalid device."""
        api = MasterthermAPI(
            VALID_LOGIN["uname"], VALID_LOGIN["upwd"], self.client, api_version="v2"
        )
        assert await api.connect() is not {}

        with pytest.raises(MasterthermResponseFormatError):
            await api.get_device_info("1234", "2")

    async def test_offline_initial(self):
        """Test getting data when the HP is offline on initial connect."""
        api = MasterthermAPI(
            VALID_LOGIN["uname"], VALID_LOGIN["upwd"], self.client, api_version="v2"
        )
        assert await api.connect() is not {}

        self.error_type = "unavailable"
        try:
            await api.get_device_data("10021", "1")
        except MasterthermPumpError as ex:
            assert ex.status == MasterthermPumpError.OFFLINE

    async def test_offline_update(self):
        """Test getting data when the HP is offline on update."""
        api = MasterthermAPI(
            VALID_LOGIN["uname"], VALID_LOGIN["upwd"], self.client, api_version="v2"
        )
        assert await api.connect() is not {}

        data = await api.get_device_data("10021", "1")
        assert data != {}

        self.error_type = "unavailable"
        try:
            await api.get_device_data("10021", "1", last_update_time=None)
        except MasterthermPumpError as ex:
            assert ex.status == MasterthermPumpError.OFFLINE

    async def test_setdata(self):
        """Test we can send data to update the API."""
        api = MasterthermAPI(
            VALID_LOGIN["uname"], VALID_LOGIN["upwd"], self.client, api_version="v2"
        )
        assert await api.connect() is not {}

        # Read the Data and issue, Power State
        data = await api.get_device_data("10021", "1")
        assert data["data"]["varData"]["001"]["D_3"] == "1"

        # Issue an update back to the API
        assert await api.set_device_data("10021", "1", "D_3", "0")

        # Re-Read to check the update is correct.
        data = await api.get_device_data("10021", "1")
        assert data["data"]["varData"]["001"]["D_3"] == "0"

    async def test_unsupported_version(self):
        """Test exception for unsupported version."""
        with pytest.raises(MasterthermUnsupportedVersion):
            MasterthermAPI("1234", "1", self.client, api_version="v10")

    async def test_expired_token(self):
        """Test for an expired token and re-connect."""
        api = MasterthermAPI(
            VALID_LOGIN["uname"], VALID_LOGIN["upwd"], self.client, api_version="v2"
        )
        self.error_type = "token_expire"
        assert await api.connect() is not {}

        assert await api.get_device_info("10021", "1")
        assert await api.get_device_data("10021", "1")
        assert await api.set_device_data("10021", "1", "D_3", "0")

    async def test_client_connection_error(self):
        """Test for Client Connection Errors, such as timeout."""
        api = MasterthermAPI(
            VALID_LOGIN["uname"], VALID_LOGIN["upwd"], self.client, api_version="v2"
        )
        assert await api.connect()
        assert await api.get_device_info("10021", "1")

        self.error_type = "ClientConnectionError"
        with pytest.raises(MasterthermConnectionError):
            await api.connect()

        with pytest.raises(MasterthermConnectionError):
            await api.get_device_info("10021", "1")

        with pytest.raises(MasterthermConnectionError):
            await api.set_device_data("10021", "1", "D_3", "0")



"""Test the Initialization of the API."""
from unittest.mock import patch

from aiohttp import ClientSession
import pytest

from masterthermconnect import (
    MasterthermController,
    MasterthermAuthenticationError,
    MasterthermConnectionError,
    MasterthermEntryNotFound,
    MasterthermPumpError,
)
from masterthermconnect.const import URL_BASE

from .conftest import VALID_LOGIN, ConnectionMock


async def test_setup():
    """Test the Controller Sets up correctly."""
    controller = MasterthermController(
        VALID_LOGIN["uname"], VALID_LOGIN["upwd"], ClientSession()
    )
    assert controller is not None


async def test_connect():
    """Test the Controller Connects and setup devices."""
    controller = MasterthermController(
        VALID_LOGIN["uname"], VALID_LOGIN["upwd"], ClientSession()
    )
    mockconnect = ConnectionMock()

    with patch(
        "masterthermconnect.api.MasterthermAPI.connect",
        return_value=mockconnect.connect(),
    ) as mock_apiconnect:
        assert await controller.connect() is True

    assert len(mock_apiconnect.mock_calls) > 0


async def test_connect_failure():
    """Test the Controller Invalid Login Connection."""
    controller = MasterthermController(
        VALID_LOGIN["uname"], VALID_LOGIN["upwd"], ClientSession()
    )

    with patch(
        "masterthermconnect.api.MasterthermAPI.connect",
        side_effect=MasterthermAuthenticationError(
            "1", "Invalid user name or password"
        ),
    ) as mock_apiconnect, pytest.raises(MasterthermAuthenticationError):
        await controller.connect()

    assert len(mock_apiconnect.mock_calls) > 0


async def test_connect_error():
    """Test the Controller on Connection Failure."""
    controller = MasterthermController(
        VALID_LOGIN["uname"], VALID_LOGIN["upwd"], ClientSession()
    )

    with patch(
        "masterthermconnect.api.MasterthermAPI.connect",
        side_effect=MasterthermConnectionError("500", "Some Other Error"),
    ) as mock_apiconnect, pytest.raises(MasterthermConnectionError):
        await controller.connect()

    assert len(mock_apiconnect.mock_calls) > 0


async def test_get_info_data():
    """Test the Controller Connects and setup devices."""
    controller = MasterthermController(
        VALID_LOGIN["uname"], VALID_LOGIN["upwd"], ClientSession()
    )
    mockconnect = ConnectionMock()

    with patch(
        "masterthermconnect.api.MasterthermAPI.connect",
        return_value=mockconnect.connect(),
    ) as mock_api_connect, patch(
        "masterthermconnect.api.MasterthermAPI.get_device_info",
        side_effect=mockconnect.get_device_info,
    ) as mock_get_device_info, patch(
        "masterthermconnect.api.MasterthermAPI.get_device_data",
        side_effect=mockconnect.get_device_data,
    ) as mock_get_device_data:
        assert await controller.connect() is True
        assert await controller.refresh() is True

    assert len(mock_api_connect.mock_calls) > 0
    assert len(mock_get_device_info.mock_calls) > 0
    assert len(mock_get_device_data.mock_calls) > 0

    assert controller.get_devices()

    info = controller.get_device_info("1234", "1")
    data_raw = controller.get_device_registers("1234", "1")
    data = controller.get_device_data("1234", "1")

    assert info["country"] == "UK"
    assert data_raw["A_500"] == "32.1"
    assert data["hp_power_state"] is True
    assert data["outside_temp"] == 5.3
    assert data["actual_temp"] == 32.1
    assert data["heating_circuits"]["hc1"]["name"] == "HW-ANN"
    assert data["heating_circuits"]["hc1"]["on"] is False
    assert "hc3" not in data["heating_circuits"]
    assert "pool" not in data["heating_circuits"]
    assert "solar" not in data["heating_circuits"]


async def test_pool_solar():
    """Test the Controller Gets Pool and Solar."""
    controller = MasterthermController(
        VALID_LOGIN["uname"], VALID_LOGIN["upwd"], ClientSession()
    )
    mockconnect = ConnectionMock(api_version="v1", use_mt=True)

    with patch(
        "masterthermconnect.api.MasterthermAPI.connect",
        return_value=mockconnect.connect(),
    ) as mock_api_connect, patch(
        "masterthermconnect.api.MasterthermAPI.get_device_info",
        side_effect=mockconnect.get_device_info,
    ) as mock_get_device_info, patch(
        "masterthermconnect.api.MasterthermAPI.get_device_data",
        side_effect=mockconnect.get_device_data,
    ) as mock_get_device_data:
        assert await controller.connect() is True
        assert await controller.refresh() is True

    assert len(mock_api_connect.mock_calls) > 0
    assert len(mock_get_device_info.mock_calls) > 0
    assert len(mock_get_device_data.mock_calls) > 0

    assert controller.get_devices()

    info = controller.get_device_info("0001", "1")
    data = controller.get_device_data("0001", "1")

    assert info["country"] == "DE"
    assert data["heating_circuits"]["pool"]
    assert data["heating_circuits"]["solar"]


async def test_operating_mode_idle():
    """Test the Controller Operating Mode."""
    controller = MasterthermController(
        VALID_LOGIN["uname"], VALID_LOGIN["upwd"], ClientSession()
    )
    mockconnect = ConnectionMock(api_version="v1", use_mt=True)

    with patch(
        "masterthermconnect.api.MasterthermAPI.connect",
        return_value=mockconnect.connect(),
    ) as mock_api_connect, patch(
        "masterthermconnect.api.MasterthermAPI.get_device_info",
        side_effect=mockconnect.get_device_info,
    ) as mock_get_device_info, patch(
        "masterthermconnect.api.MasterthermAPI.get_device_data",
        side_effect=mockconnect.get_device_data,
    ) as mock_get_device_data:
        assert await controller.connect() is True
        assert await controller.refresh() is True

    assert len(mock_api_connect.mock_calls) > 0
    assert len(mock_get_device_info.mock_calls) > 0
    assert len(mock_get_device_data.mock_calls) > 0

    data = controller.get_device_data("0524", "4")

    assert data["operating_mode"] == "idle"


async def test_operating_mode_cooling():
    """Test the Passive Cooling Operating Mode."""
    controller = MasterthermController(
        VALID_LOGIN["uname"], VALID_LOGIN["upwd"], ClientSession()
    )
    mockconnect = ConnectionMock(api_version="v2")

    with patch(
        "masterthermconnect.api.MasterthermAPI.connect",
        return_value=mockconnect.connect(),
    ), patch(
        "masterthermconnect.api.MasterthermAPI.get_device_info",
        side_effect=mockconnect.get_device_info,
    ), patch(
        "masterthermconnect.api.MasterthermAPI.get_device_data",
        side_effect=mockconnect.get_device_data,
    ):
        assert await controller.connect() is True
        assert await controller.refresh() is True

    data = controller.get_device_data("10021", "2")

    assert data["operating_mode"] == "cooling"
    assert data["requested_temp"] == 15.1


async def test_operating_mode_heating():
    """Test the Controller Operating Mode."""
    controller = MasterthermController(
        VALID_LOGIN["uname"], VALID_LOGIN["upwd"], ClientSession()
    )
    mockconnect = ConnectionMock()

    with patch(
        "masterthermconnect.api.MasterthermAPI.connect",
        return_value=mockconnect.connect(),
    ) as mock_api_connect, patch(
        "masterthermconnect.api.MasterthermAPI.get_device_info",
        side_effect=mockconnect.get_device_info,
    ) as mock_get_device_info, patch(
        "masterthermconnect.api.MasterthermAPI.get_device_data",
        side_effect=mockconnect.get_device_data,
    ) as mock_get_device_data:
        assert await controller.connect() is True
        assert await controller.refresh() is True

    assert len(mock_api_connect.mock_calls) > 0
    assert len(mock_get_device_info.mock_calls) > 0
    assert len(mock_get_device_data.mock_calls) > 0

    data = controller.get_device_data("1234", "1")

    assert data["operating_mode"] == "heating"
    assert data["requested_temp"] == 34.7


async def test_operating_mode_dhw():
    """Test the Controller Operating Mode DHW."""
    controller = MasterthermController(
        VALID_LOGIN["uname"], VALID_LOGIN["upwd"], ClientSession()
    )
    mockconnect = ConnectionMock(api_version="v1", use_mt=True)

    with patch(
        "masterthermconnect.api.MasterthermAPI.connect",
        return_value=mockconnect.connect(),
    ) as mock_api_connect, patch(
        "masterthermconnect.api.MasterthermAPI.get_device_info",
        side_effect=mockconnect.get_device_info,
    ) as mock_get_device_info, patch(
        "masterthermconnect.api.MasterthermAPI.get_device_data",
        side_effect=mockconnect.get_device_data,
    ) as mock_get_device_data:
        assert await controller.connect() is True
        assert await controller.refresh() is True

    assert len(mock_api_connect.mock_calls) > 0
    assert len(mock_get_device_info.mock_calls) > 0
    assert len(mock_get_device_data.mock_calls) > 0

    data = controller.get_device_data("0002", "1")

    assert data["operating_mode"] == "dhw"
    assert data["requested_temp"] == 45.0


async def test_new_api_get_info_data():
    """Test the Controller Connects and setup devices for the New API."""
    controller = MasterthermController(
        VALID_LOGIN["uname"], VALID_LOGIN["upwd"], ClientSession()
    )
    mockconnect = ConnectionMock(api_version="v2")

    with patch(
        "masterthermconnect.api.MasterthermAPI.connect",
        return_value=mockconnect.connect(),
    ) as mock_api_connect, patch(
        "masterthermconnect.api.MasterthermAPI.get_device_info",
        side_effect=mockconnect.get_device_info,
    ) as mock_get_device_info, patch(
        "masterthermconnect.api.MasterthermAPI.get_device_data",
        side_effect=mockconnect.get_device_data,
    ) as mock_get_device_data:
        assert await controller.connect() is True
        assert await controller.refresh() is True

    assert len(mock_api_connect.mock_calls) > 0
    assert len(mock_get_device_info.mock_calls) > 0
    assert len(mock_get_device_data.mock_calls) > 0

    assert controller.get_devices()

    info = controller.get_device_info("10021", "1")
    data_raw = controller.get_device_registers("10021", "1")
    data = controller.get_device_data("10021", "1")

    assert info["country"] == "CZ"
    assert info["api_url"] == URL_BASE

    assert data_raw["A_500"] == "30.5"
    assert data["hp_power_state"] is True
    assert data["outside_temp"] == 6.4
    assert data["actual_temp"] == 30.5

    assert "hc0" not in data["heating_circuits"]

    assert data["heating_circuits"]["hc1"]["name"] == "Living room"
    assert data["heating_circuits"]["hc1"]["on"] is True

    assert data["heating_circuits"]["hc2"]["name"] == "Bedroom"
    assert data["heating_circuits"]["hc2"]["on"] is False


async def test_getdata_update():
    """Test getting the data and getting an update.

    Test the Controller Connects and setup devices.
    """
    controller = MasterthermController(
        VALID_LOGIN["uname"], VALID_LOGIN["upwd"], ClientSession()
    )
    mockconnect = ConnectionMock()

    with patch(
        "masterthermconnect.api.MasterthermAPI.connect",
        return_value=mockconnect.connect(),
    ) as mock_apiconnect, patch(
        "masterthermconnect.api.MasterthermAPI.get_device_info",
        side_effect=mockconnect.get_device_info,
    ) as mock_get_device_info, patch(
        "masterthermconnect.api.MasterthermAPI.get_device_data",
        side_effect=mockconnect.get_device_data,
    ) as mock_get_device_data:
        assert await controller.connect() is True
        controller.set_refresh_rate(data_refresh_seconds=0, data_offset_seconds=0)
        assert await controller.refresh() is True

        data = controller.get_device_data("1234", "1")
        assert data["actual_temp"] == 32.1
        assert await controller.refresh() is True

    assert len(mock_apiconnect.mock_calls) == 1
    assert len(mock_get_device_info.mock_calls) == 1
    assert len(mock_get_device_data.mock_calls) == 2

    assert controller.get_devices()

    data_raw = controller.get_device_registers("1234", "1")
    data_update = controller.get_device_registers("1234", "1", last_updated=True)
    data = controller.get_device_data("1234", "1")

    assert data_raw["A_500"] == "40.7"
    assert data_update["A_500"] == "40.7"
    assert "I_418" not in data_update
    assert "I_418" in data_raw
    assert data["actual_temp"] == 40.7


async def test_periodic_full_load():
    """Test getting full data on refresh time."""
    controller = MasterthermController(
        VALID_LOGIN["uname"], VALID_LOGIN["upwd"], ClientSession()
    )
    mockconnect = ConnectionMock()

    with patch(
        "masterthermconnect.api.MasterthermAPI.connect",
        return_value=mockconnect.connect(),
    ) as mock_apiconnect, patch(
        "masterthermconnect.api.MasterthermAPI.get_device_info",
        side_effect=mockconnect.get_device_info,
    ) as mock_get_device_info, patch(
        "masterthermconnect.api.MasterthermAPI.get_device_data",
        side_effect=mockconnect.get_device_data,
    ) as mock_get_device_data:
        assert await controller.connect() is True
        controller.set_refresh_rate(
            data_refresh_seconds=0, data_offset_seconds=0, full_refresh_minutes=0
        )
        assert await controller.refresh() is True

        data = controller.get_device_data("1234", "1")
        assert data["actual_temp"] == 32.1
        assert await controller.refresh() is True

    assert len(mock_apiconnect.mock_calls) == 1
    assert len(mock_get_device_info.mock_calls) == 1
    assert len(mock_get_device_data.mock_calls) == 2


async def test_season_winter():
    """Test the Controller Season."""
    controller = MasterthermController(
        VALID_LOGIN["uname"], VALID_LOGIN["upwd"], ClientSession()
    )
    mockconnect = ConnectionMock()

    with patch(
        "masterthermconnect.api.MasterthermAPI.connect",
        return_value=mockconnect.connect(),
    ) as mock_api_connect, patch(
        "masterthermconnect.api.MasterthermAPI.get_device_info",
        side_effect=mockconnect.get_device_info,
    ) as mock_get_device_info, patch(
        "masterthermconnect.api.MasterthermAPI.get_device_data",
        side_effect=mockconnect.get_device_data,
    ) as mock_get_device_data:
        assert await controller.connect() is True
        assert await controller.refresh() is True

    assert len(mock_api_connect.mock_calls) > 0
    assert len(mock_get_device_info.mock_calls) > 0
    assert len(mock_get_device_data.mock_calls) > 0

    data = controller.get_device_data("1234", "1")

    assert data["season"]["mode"] == "winter"


async def test_season_auto_winter():
    """Test the Controller Season Auto Winter."""
    controller = MasterthermController(
        VALID_LOGIN["uname"], VALID_LOGIN["upwd"], ClientSession()
    )
    mockconnect = ConnectionMock(api_version="v1", use_mt=True)

    with patch(
        "masterthermconnect.api.MasterthermAPI.connect",
        return_value=mockconnect.connect(),
    ) as mock_api_connect, patch(
        "masterthermconnect.api.MasterthermAPI.get_device_info",
        side_effect=mockconnect.get_device_info,
    ) as mock_get_device_info, patch(
        "masterthermconnect.api.MasterthermAPI.get_device_data",
        side_effect=mockconnect.get_device_data,
    ) as mock_get_device_data:
        assert await controller.connect() is True
        assert await controller.refresh() is True

    assert len(mock_api_connect.mock_calls) > 0
    assert len(mock_get_device_info.mock_calls) > 0
    assert len(mock_get_device_data.mock_calls) > 0

    data = controller.get_device_data("0001", "1")

    assert data["season"]["mode"] == "winter"


async def test_toggle_hp_on():
    """Test toggling the HP Main Switch."""
    controller = MasterthermController(
        VALID_LOGIN["uname"], VALID_LOGIN["upwd"], ClientSession()
    )
    mockconnect = ConnectionMock(api_version="v1", use_mt=True)

    with patch(
        "masterthermconnect.api.MasterthermAPI.connect",
        return_value=mockconnect.connect(),
    ), patch(
        "masterthermconnect.api.MasterthermAPI.get_device_info",
        side_effect=mockconnect.get_device_info,
    ), patch(
        "masterthermconnect.api.MasterthermAPI.get_device_data",
        side_effect=mockconnect.get_device_data,
    ), patch(
        "masterthermconnect.api.MasterthermAPI.set_device_data",
        side_effect=mockconnect.set_device_data,
    ) as mock_set:
        assert await controller.connect() is True
        assert await controller.refresh() is True

        data = controller.get_device_data("0001", "1")
        state = data["hp_power_state"]

        assert await controller.set_device_data_item(
            "0001", "1", "hp_power_state", not state
        )

        await controller.refresh(full_load=True)
        data = controller.get_device_data_item("0001", "1", "hp_power_state")
        assert data != state

    assert len(mock_set.mock_calls) > 0


async def test_toggle_hc1_on():
    """Test toggling the Heating Circuit 1 Switch."""
    controller = MasterthermController(
        VALID_LOGIN["uname"], VALID_LOGIN["upwd"], ClientSession()
    )
    mockconnect = ConnectionMock(api_version="v1")

    with patch(
        "masterthermconnect.api.MasterthermAPI.connect",
        return_value=mockconnect.connect(),
    ), patch(
        "masterthermconnect.api.MasterthermAPI.get_device_info",
        side_effect=mockconnect.get_device_info,
    ), patch(
        "masterthermconnect.api.MasterthermAPI.get_device_data",
        side_effect=mockconnect.get_device_data,
    ), patch(
        "masterthermconnect.api.MasterthermAPI.set_device_data",
        side_effect=mockconnect.set_device_data,
    ) as mock_set:
        assert await controller.connect() is True
        assert await controller.refresh() is True

        data = controller.get_device_data("1234", "1")
        state = data["heating_circuits"]["hc1"]["on"]

        assert await controller.set_device_data_item(
            "1234", "1", "heating_circuits.hc1.on", not state
        )

        await controller.refresh(full_load=True)
        data = controller.get_device_data_item("1234", "1", "heating_circuits.hc1.on")
        assert data != state

    assert len(mock_set.mock_calls) > 0


async def test_set_not_valid():
    """Test that value that can't be updated is disabled."""
    controller = MasterthermController(
        VALID_LOGIN["uname"], VALID_LOGIN["upwd"], ClientSession()
    )
    mockconnect = ConnectionMock(api_version="v1")

    with patch(
        "masterthermconnect.api.MasterthermAPI.connect",
        return_value=mockconnect.connect(),
    ), patch(
        "masterthermconnect.api.MasterthermAPI.get_device_info",
        side_effect=mockconnect.get_device_info,
    ), patch(
        "masterthermconnect.api.MasterthermAPI.get_device_data",
        side_effect=mockconnect.get_device_data,
    ), patch(
        "masterthermconnect.api.MasterthermAPI.set_device_data",
        side_effect=mockconnect.set_device_data,
    ):
        assert await controller.connect() is True
        assert await controller.refresh() is True

        data = controller.get_device_data("1234", "1")
        temp = data["heating_circuits"]["hc1"]["ambient_temp"]

        with pytest.raises(MasterthermEntryNotFound):
            await controller.set_device_data_item(
                "1234", "1", "heating_circuits.hc1.ambient_temp", 50.1
            )

        await controller.refresh(full_load=True)
        data = controller.get_device_data_item(
            "1234", "1", "heating_circuits.hc1.ambient_temp"
        )
        assert data == temp


async def test_cooling_feature():
    """Test the Controller Correctly disables Cooling Mode."""
    controller = MasterthermController(
        VALID_LOGIN["uname"], VALID_LOGIN["upwd"], ClientSession()
    )
    mockconnect = ConnectionMock(api_version="v2")

    with patch(
        "masterthermconnect.api.MasterthermAPI.connect",
        return_value=mockconnect.connect(),
    ), patch(
        "masterthermconnect.api.MasterthermAPI.get_device_info",
        side_effect=mockconnect.get_device_info,
    ), patch(
        "masterthermconnect.api.MasterthermAPI.get_device_data",
        side_effect=mockconnect.get_device_data,
    ):
        assert await controller.connect() is True
        assert await controller.refresh() is True

    info1 = controller.get_device_info("10021", "1")
    data1 = controller.get_device_data("10021", "1")
    info2 = controller.get_device_info("10021", "2")
    data2 = controller.get_device_data("10021", "2")

    # Unit 1 has cooling disabled
    assert info1["cooling"] == "0"
    assert "hp_function" not in data1
    assert "control_curve_cooling" not in data1

    # Unit 2 has cooling enabled
    assert info2["cooling"] == "1"
    assert "hp_function" in data2
    assert "control_curve_cooling" in data2


async def test_set_cooling_feature():
    """Test cooling feature setting."""
    controller = MasterthermController(
        VALID_LOGIN["uname"], VALID_LOGIN["upwd"], ClientSession()
    )
    mockconnect = ConnectionMock(api_version="v2")

    with patch(
        "masterthermconnect.api.MasterthermAPI.connect",
        return_value=mockconnect.connect(),
    ), patch(
        "masterthermconnect.api.MasterthermAPI.get_device_info",
        side_effect=mockconnect.get_device_info,
    ), patch(
        "masterthermconnect.api.MasterthermAPI.get_device_data",
        side_effect=mockconnect.get_device_data,
    ), patch(
        "masterthermconnect.api.MasterthermAPI.set_device_data",
        side_effect=mockconnect.set_device_data,
    ):
        assert await controller.connect() is True
        assert await controller.refresh() is True

        with pytest.raises(MasterthermEntryNotFound):
            await controller.set_device_data_item("10021", "1", "hp_function", 1)

        with pytest.raises(MasterthermEntryNotFound):
            await controller.set_device_data_item(
                "10021", "1", "control_curve_cooling.setpoint_a_outside", 20.2
            )

        await controller.set_device_data_item("10021", "2", "hp_function", 1)
        await controller.set_device_data_item(
            "10021", "2", "control_curve_cooling.setpoint_a_outside", 20.2
        )


async def test_ground_source():
    """Test the HP Types and what type of thermal pump."""
    controller = MasterthermController(
        VALID_LOGIN["uname"], VALID_LOGIN["upwd"], ClientSession()
    )
    mockconnect = ConnectionMock(api_version="v1")

    with patch(
        "masterthermconnect.api.MasterthermAPI.connect",
        return_value=mockconnect.connect(),
    ), patch(
        "masterthermconnect.api.MasterthermAPI.get_device_info",
        side_effect=mockconnect.get_device_info,
    ), patch(
        "masterthermconnect.api.MasterthermAPI.get_device_data",
        side_effect=mockconnect.get_device_data,
    ):
        assert await controller.connect() is True
        assert await controller.refresh() is True

    assert controller.get_device_data_item("1234", "1", "hp_type") == 1
    assert "brine_pump_running" in controller.get_device_data("1234", "1")


async def test_air_source():
    """Test the HP Types and what type of thermal pump."""
    controller = MasterthermController(
        VALID_LOGIN["uname"], VALID_LOGIN["upwd"], ClientSession()
    )
    mockconnect = ConnectionMock(api_version="v1", use_mt=True)

    with patch(
        "masterthermconnect.api.MasterthermAPI.connect",
        return_value=mockconnect.connect(),
    ), patch(
        "masterthermconnect.api.MasterthermAPI.get_device_info",
        side_effect=mockconnect.get_device_info,
    ), patch(
        "masterthermconnect.api.MasterthermAPI.get_device_data",
        side_effect=mockconnect.get_device_data,
    ):
        assert await controller.connect() is True
        assert await controller.refresh() is True

    data = controller.get_device_data("0001", "1")
    assert data["hp_type"] == 4
    assert "fan_running" in data

    data = controller.get_device_data("0524", "1")
    assert data["hp_type"] == 0
    assert "fan_running" in data


async def test_pump_offline_connect():
    """Test the Pump being offline on initial connect."""
    controller = MasterthermController(
        VALID_LOGIN["uname"], VALID_LOGIN["upwd"], ClientSession()
    )
    mockconnect = ConnectionMock(api_version="v1")

    with patch(
        "masterthermconnect.api.MasterthermAPI.connect",
        return_value=mockconnect.connect(),
    ), patch(
        "masterthermconnect.api.MasterthermAPI.get_device_info",
        side_effect=mockconnect.get_device_info,
    ), patch(
        "masterthermconnect.api.MasterthermAPI.get_device_data",
        side_effect=MasterthermPumpError(
            MasterthermPumpError.OFFLINE,
            "Actual data for some variables are not available.",
        ),
    ):
        assert await controller.connect() is True
        assert await controller.refresh() is True

    assert controller.get_device_data_item("1234", "1", "operating_mode") == "offline"


async def test_pump_offline_update():
    """Test the Pump being offline after connect on update."""
    controller = MasterthermController(
        VALID_LOGIN["uname"], VALID_LOGIN["upwd"], ClientSession()
    )
    mockconnect = ConnectionMock(api_version="v1")

    with patch(
        "masterthermconnect.api.MasterthermAPI.connect",
        return_value=mockconnect.connect(),
    ), patch(
        "masterthermconnect.api.MasterthermAPI.get_device_info",
        side_effect=mockconnect.get_device_info,
    ), patch(
        "masterthermconnect.api.MasterthermAPI.get_device_data",
        side_effect=mockconnect.get_device_data,
    ):
        assert await controller.connect() is True
        assert await controller.refresh() is True

    with patch(
        "masterthermconnect.api.MasterthermAPI.get_device_info",
        side_effect=mockconnect.get_device_info,
    ), patch(
        "masterthermconnect.api.MasterthermAPI.get_device_data",
        side_effect=MasterthermPumpError(
            MasterthermPumpError.OFFLINE,
            "Actual data for some variables are not available.",
        ),
    ):
        assert await controller.refresh(full_load=True)

    assert controller.get_device_data_item("1234", "1", "operating_mode") == "offline"


async def test_diagnostics_data():
    """Test the Diagnostics Data comes back correctly."""
    controller = MasterthermController(
        VALID_LOGIN["uname"], VALID_LOGIN["upwd"], ClientSession()
    )
    mockconnect = ConnectionMock(api_version="v2")

    with patch(
        "masterthermconnect.api.MasterthermAPI.connect",
        return_value=mockconnect.connect(),
    ), patch(
        "masterthermconnect.api.MasterthermAPI.get_device_info",
        side_effect=mockconnect.get_device_info,
    ), patch(
        "masterthermconnect.api.MasterthermAPI.get_device_data",
        side_effect=mockconnect.get_device_data,
    ):
        assert await controller.connect() is True
        assert await controller.refresh() is True

        diag_data = controller.get_diagnostics_data()
        # Verify Sensitive Data:

        # Info Sensitive
        info = controller.get_device_info("10021", "1")
        diag_info = diag_data["1112_1"]["info"]
        assert diag_info["module_id"] != info["module_id"]
        assert diag_info["module_name"] != info["module_name"]
        assert diag_info["name"] != info["name"]
        assert diag_info["surname"] != info["surname"]
        assert diag_info["latitude"] != info["latitude"]
        assert diag_info["longitude"] != info["longitude"]
        assert diag_info["place"] != info["place"]
        assert diag_info["notes"] != info["notes"]


async def test_diagnostics_nohide():
    """Test the Diagnostics Data comes back correctly."""
    controller = MasterthermController(
        VALID_LOGIN["uname"], VALID_LOGIN["upwd"], ClientSession()
    )
    mockconnect = ConnectionMock(api_version="v1")

    with patch(
        "masterthermconnect.api.MasterthermAPI.connect",
        return_value=mockconnect.connect(),
    ), patch(
        "masterthermconnect.api.MasterthermAPI.get_device_info",
        side_effect=mockconnect.get_device_info,
    ), patch(
        "masterthermconnect.api.MasterthermAPI.get_device_data",
        side_effect=mockconnect.get_device_data,
    ):
        assert await controller.connect() is True
        assert await controller.refresh() is True

        diag_data = controller.get_diagnostics_data(hide_sensitive=False)
        # Verify Sensitive Data:

        info = controller.get_device_info("1234", "1")
        diag_info = diag_data["1234_1"]["info"]
        assert diag_info["module_id"] == info["module_id"]
        assert diag_info["module_name"] == info["module_name"]
        assert diag_info["name"] == info["name"]
        assert diag_info["surname"] == info["surname"]
        assert diag_info["latitude"] == info["latitude"]
        assert diag_info["longitude"] == info["longitude"]
        assert diag_info["place"] == info["place"]
        assert diag_info["notes"] == info["notes"]



"""Special Class for use with the Mapping Files."""


class Special:
    """A Special Class Type for use in mappings."""

    FIXED = "fixed"
    NAMEARRAY = "namearray"
    FORMULA = "formula"

    def __init__(self, data_type: any, condition: str):
        """Initialise the Special Class with the Condition Needed."""
        self.data_type = data_type
        self.condition = condition

    def evaluate(self, formula: str, values: list) -> any:
        """Evalulate the items based on the condition."""
        converted_formula = formula.format(*values)
        result = eval(converted_formula, {}, {})  # pylint: disable=W0123
        return result
